##############################################################################
#
# Copyright (c) 2002, 2005 Nexedi SARL and Contributors. All Rights Reserved.
#                    Sebastien Robin <seb@nexedi.com>
#                    Jean-Paul Smets <jp@nexedi.com>
#                    Romain Courteaud <romain@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from Constraint import Constraint
from DateTime import DateTime

class PropertyTypeValidity(Constraint):
  """
    This constraint class allows to check / fix type of each 
    attributes define in the PropertySheets.
    This Constraint is always created in ERP5Type/Utils.py
  """

  # Initialize type dict
  _type_dict = {
    'string':             (type('a'), ),
    'text':               (type('a'), ),
    'int':                (type(1), ),
    'boolean':            (type(1), ),
    'float':              (type(1.0), ),
    'long':               (type(1L), ),
    'tales':              (type('string:3'), ),
    'lines':              (type([]), type(())),
    'tokens':             (type([]), type(())),
    'selection':          (type([]), type(())),
    'multiple selection': (type([]), type(())),
    'date':               (type(DateTime()), ),
  }

  def checkConsistency(self, object, fixit=0):
    """
      This is the check method, we return a list of string,
      each string corresponds to an error.
    """
    errors = []
    # For each attribute name, we check type
    for property in object.propertyMap():
      property_id = property['id']
      property_type = property['type']
      wrong_type = 0
      value = object.getProperty(property_id)
      if value is not None:
        # Check known type
        try:
          wrong_type = (type(value) not in self._type_dict[property_type])
        except KeyError:
          wrong_type = 0
          error_message = "Attribute %s is defined with unknown type %s" % \
                          (property_id, property_type)
          errors.append(self._generateError(object, error_message))
      if wrong_type:
        # Type is wrong, so, raise constraint error
        error_message = \
            "Attribute %s should be of type %s but is of type %s" % \
            (property_id, property_type, str(type(value)))
        if fixit:
          # try to cast to correct type
          if wrong_type :
            try :
              value = self._type_dict[property_type][0](value)
            except (KeyError, ValueError), error:
              error_message += " (Type cast failed : %s)" % error
          object.setProperty(property_id, value)
          error_message += " (Fixed)"
        errors.append(self._generateError(object, error_message))
    return errors
